@@ -1,20 +1,22 @@
 import RPi.GPIO as GPIO
 import os
+import copy
 import time
 from datetime import datetime, timedelta
 from pathlib import Path
 from baseutils_phornee import ManagedClass
+from baseutils_phornee import Logger
 from influxdb_client import InfluxDBClient, Point, WritePrecision
 from influxdb_client.client.write_api import SYNCHRONOUS
 import json
 
-min_date = datetime(1971, 11, 24, 0, 0, 0)
-
 class Waterflow(ManagedClass):
 
     def __init__(self):
         super().__init__(execpath=__file__)
 
+        self.logger = Logger({'modulename': self.getClassName(), 'logpath': 'log'})
+
         token = self.config['influxdbconn']['token']
         self.org = self.config['influxdbconn']['org']
         self.bucket = self.config['influxdbconn']['bucket']
@@ -25,13 +27,6 @@ class Waterflow(ManagedClass):
     def getClassName(cls):
         return "waterflow"
 
-    @classmethod
-    def getLog(cls):
-        log_path = os.path.join(cls.getHomevarPath(), 'log/waterflow.log')
-
-        with open(log_path, 'r') as file:
-            return file.read()
-
     def readConfig(self):
         super().readConfig()
 
@@ -42,6 +37,17 @@ class Waterflow(ManagedClass):
         # Sort the programs by time
         self.config['programs'].sort(key=lambda prog: prog['start_time'])
 
+    def writeConfig(self, config):
+        conf = copy.deepcopy(config)
+
+        #self.setCache(config)
+
+        # Convert the date back from datetime to string
+        for program in conf['programs']:
+            program['start_time'] = program['start_time'].strftime('%H:%M:%S')
+
+        super().writeConfig(conf)
+
     def _timeToStr(self, time_var):
         return time_var.strftime('%Y-%m-%d %H:%M:%S\n')
 
@@ -59,17 +65,21 @@ class Waterflow(ManagedClass):
 
 
     def _recalcNextProgram(self, last_program_time):
+        """
+        Calculates which is the next program to be executed
+        """
         next_program_time = None
         program_number = -1
 
         current_time = datetime.now().replace(microsecond=0)
 
-        # Find if next program is today
+        # Find if next program is today, considering the last program time executed
         for idx, program in enumerate(self.config['programs']):
             if program['enabled'] == True:
                 candidate_time = current_time.replace(hour=program['start_time'].hour,
                                                       minute=program['start_time'].minute,
                                                       second=0)
+                # If this candidate is after the last one executed AND its no more that 10 minutes in the past, choose it
                 if candidate_time > last_program_time:
                     next_program_time = candidate_time
                     program_number = idx
@@ -112,6 +122,7 @@ class Waterflow(ManagedClass):
 
     def getLock(self):
         """
+        This is to ensure that only one execution will run from cron at the same time
         Use file as a lock... not using DB locks because we want to maximize resiliency
         """
         lock_path = os.path.join(self.homevar, 'lock')
@@ -123,6 +134,7 @@ class Waterflow(ManagedClass):
             modified_time = datetime.fromtimestamp(os.path.getmtime(lock_path))
             if (datetime.utcnow() - modified_time) > timedelta(minutes=20):
                 self.logger.warning('Lock expired: Last loop ended abnormally?.')
+                Path(lock_path).touch() # Previous token expired (previous loop crashed?)... so we will retouch to try again
                 return True
         return False
 
@@ -134,46 +146,40 @@ class Waterflow(ManagedClass):
         else:
             self.logger.error(f"Could not release lock.")
 
-    @classmethod
-    def isLoopingCorrectly(cls):
-        tokenpath = os.path.join(cls.getHomevarPath(), 'token')
+    def isLoopingCorrectly(self):
+        tokenpath = os.path.join(self.getHomevarPath(), 'token')
 
         modTimesinceEpoc = os.path.getmtime(tokenpath)
         modificationTime = datetime.utcfromtimestamp(modTimesinceEpoc)
 
         return (datetime.utcnow() - modificationTime) < timedelta(minutes=10)
 
-    @classmethod
-    def force(cls, type_force, value):
-        config = cls.getConfig()
+    def force(self, type_force, value):
+        config = self.getConfig()
         if (type_force == 'program' and 0 <= value < len(config['programs'])) or \
            (type_force == 'valve' and 0 <= value < len(config['valves'])):
-            force_file_path = os.path.join(cls.getHomevarPath(), 'force')
+            force_file_path = os.path.join(self.getHomevarPath(), 'force')
             with open(force_file_path, 'w') as force_file:
                 force_file.write('{{"type":"{0}","value":{1}}}'.format(type_force, value))
                 return True
         else:
             return False
 
-    @classmethod
-    def stop(cls):
-        stop_req_path = os.path.join(cls.getHomevarPath(), 'stop')
+    def stop(self):
+        stop_req_path = os.path.join(self.getHomevarPath(), 'stop')
         Path(stop_req_path).touch()
         return True
 
-    @classmethod
-    def stopRequested(cls):
-        stop_req_path = os.path.join(cls.getHomevarPath(), 'stop')
+    def stopRequested(self):
+        stop_req_path = os.path.join(self.getHomevarPath(), 'stop')
         return os.path.exists(stop_req_path)
 
-    @classmethod
-    def stopRemove(cls):
-        stop_req_path = os.path.join(cls.getHomevarPath(), 'stop')
+    def stopRemove(self):
+        stop_req_path = os.path.join(self.getHomevarPath(), 'stop')
         return os.remove(stop_req_path)
 
-    @classmethod
-    def getForcedInfo(cls):
-        force_file_path = os.path.join(cls.getHomevarPath(), 'force')
+    def getForcedInfo(self):
+        force_file_path = os.path.join(self.getHomevarPath(), 'force')
         if os.path.exists(force_file_path):
             with open(force_file_path, 'r') as force_file:
                 data = json.load(force_file)
@@ -181,13 +187,19 @@ class Waterflow(ManagedClass):
         else:
             return None
 
+    def getLog(self):
+        return self.logger.getLog()
+
     def _sleep(self, time_sleep):
+        """
+        Sleep "time_sleep" time, but checks every 5 seconds if a stop has been requested
+        """
         time_count = 0
         while not self.stopRequested() and time_count < time_sleep:
             time_count = time_count + 5
             time.sleep(5)  # Every X seconds
 
-    def _emitMetric(self, action, forced):
+    def _emitActionMetric(self, action, forced):
         write_api = self.conn.write_api(write_options=SYNCHRONOUS)
 
         point = Point('piwaterflow') \
@@ -208,7 +220,7 @@ class Waterflow(ManagedClass):
         GPIO.output(valve_pin, GPIO.HIGH)
         self.logger.info('Valve %s ON.' % valve)
 
-        self._sleep(5*60)
+        self._sleep(self.config['max_valve_time']*60)
 
         GPIO.output(valve_pin, GPIO.LOW)
         self.logger.info('Valve %s OFF.' % valve)
@@ -216,6 +228,22 @@ class Waterflow(ManagedClass):
         GPIO.output(self.config['inverter_relay_pin'], GPIO.LOW)  # INVERTER always OFF after operations
         self.logger.info('Inverter relay OFF.')
 
+    def _skipProgram(self):
+        # print(self.config['humidity_threshold'])
+        # if self.is_raspberry_pi():
+        #     import adafruit_dht
+        #     dhtSensor = adafruit_dht.DHT22(self.config['pin'])
+        #     humidity = dhtSensor.humidity
+        #     temp_c = dhtSensor.temperature
+        #     if humidity >= self.config['humidity_threshold']:
+        #         self.logger.info('(humidity {} > {}.'.format(humidity, self.config['humidity_threshold']))
+        #         return True
+        #     else:
+        #         return False
+        # else:
+        #    return False
+        return False
+
     def _executeProgram(self, program_number):
         """
         Works for regular programs, or forced ones (if program number is sent)
@@ -243,13 +271,13 @@ class Waterflow(ManagedClass):
 
     def _logNextProgramTime(self, current_time):
 
-        log = self.getLog()
+        log = self.logger.getLog()
 
         lines = log.split('\n')
         last_line = lines[-2]
 
         new_next_program_time, _ = self._recalcNextProgram(current_time)
-        if new_next_program_time is not None:
+        if new_next_program_time:
             string_to_log = 'Next program: %s.' % new_next_program_time.strftime('%Y-%m-%d %H:%M')
         else:
             string_to_log = 'NO active program!'
@@ -258,12 +286,8 @@ class Waterflow(ManagedClass):
             self.logger.info(string_to_log)
 
     def loop(self):
-        if self.getLock():  # To ensure a single execution
+        if self.getLock():  # To ensure a single execution despite of cron overlapping
             try:
-                # Updates "modified" time, so that we can keep track about waterflow looping
-                token_path = os.path.join(self.homevar, 'token')
-                Path(token_path).touch()
-
                 current_time = datetime.now()
                 forced_info = self.getForcedInfo()
 
@@ -271,46 +295,62 @@ class Waterflow(ManagedClass):
                     self._setupGPIO(self.config['valves'])
                     last_program_time = self._readLastProgramTime()
 
-                    if forced_info is not None:
+                    if forced_info:
                         forced_type = forced_info.get("type")
                         forced_value = forced_info.get("value")
                         if forced_type == "program":
                             self.logger.info('Forced program {} executing now.'.format(forced_value))
                             # ------------------------
-                            self._emitMetric('prog{}'.format(forced_value), True)
+                            self._emitActionMetric('prog{}'.format(forced_value), True)
                             self._executeProgram(forced_value)
                             self._writeLastProgramTime(self._timeToStr(current_time))
                         elif forced_type == "valve":
                             # ------------------------
-                            self._emitMetric('valve{}'.format(forced_value), True)
+                            self._emitActionMetric('valve{}'.format(forced_value), True)
                             self._executeValve(forced_value)
                     else:
                         new_next_program_time, calculated_program_number = self._recalcNextProgram(last_program_time)
-                        if new_next_program_time is not None and current_time >= new_next_program_time:
+                        if new_next_program_time:
                             # ------------------------
-                            self._emitMetric('prog{}'.format(calculated_program_number), False)
-                            self._executeProgram(calculated_program_number)
-                            self._writeLastProgramTime(self._timeToStr(current_time))
+                            time_reached = current_time >= new_next_program_time
+                            time_threshold_exceeded = current_time > (new_next_program_time + timedelta(minutes=10))
+                            skip_program = self._skipProgram()
+                            # If we have reached the time of the new_program_time, BUT not by more than 10 minutes...
+                            if time_reached and not time_threshold_exceeded and not skip_program:
+                                self._emitActionMetric('prog{}'.format(calculated_program_number), False)
+                                self._executeProgram(calculated_program_number)
+                                program_executed = True
+                            else:
+                                program_executed = False
 
-                if forced_info is not None:
+                            if program_executed or skip_program or time_threshold_exceeded:
+                                self._writeLastProgramTime(self._timeToStr(current_time))
+
+                if forced_info:
                     # Remove force token file
                     os.remove(os.path.join(self.getHomevarPath(), 'force'))
 
                 if self.stopRequested():
                     self.logger.info('Activity stopped.')
-                    self._emitMetric('Stop', True)
+                    self._emitActionMetric('Stop', True)
                     self.stopRemove()
 
                 # Recalc next program time
                 self._logNextProgramTime(current_time)
+
+                # Updates "modified" time AT THE END, so that we can keep track about waterflow looping SUCCESFULLY.
+                token_path = os.path.join(self.homevar, 'token')
+                Path(token_path).touch()
+
             except Exception as e:
-                self.logger.error(f"Exception looping: {e}")
+                self.logger.error("Exception looping: {}".format(str(e)))
+                raise
             finally:
                 GPIO.cleanup()
                 self.releaseLock()
 
-
 if __name__ == "__main__":
+
     waterflow_instance = Waterflow()
     waterflow_instance.loop()
 
